// mex intTwoK_mex.c -I/usr/local/include/ -lmwlapack -lmwblas -lgsl
// called from matlab as intTwoK_mex(u,w,x,sV)

#if !defined(_WIN32)
#define dgeqrf dgeqrf_      // for qr
#define dgemm dgemm_
#endif

#include "mex.h"
#include "lapack.h"
#include "blas.h"
#include "libc.h"
#include <math.h>
#include <gsl/gsl_integration.h>
// #include <stdlib.h>


#define DBL_EPS 2.220446049250313080847e-16

struct PARAMS {
    double * ubar;
    double * xbar;
    double * wbar;
};

double intfunc(double s,double *ubar, double *wbar, double*xbar);
double wrap_intfunc(double s,void * gsl_params);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *u,*w,*x,*sV;                 /* pointers to input matrices */
    double *Tau,*WORK,*A,*R;               /* point to working matrices */
    double *OUT;
    ptrdiff_t m;                      /* matrix dimensions and lda = m*/
    ptrdiff_t lwork = -1;               // dynamically allocate memory (for best performance)
    mwSignedIndex info;         // for function completion status
    // int info;
    ptrdiff_t rank;                   // to hold MIN(m,n)
    ptrdiff_t n = 3;
    char *TRANS = "N";
    double one = 1.0; double zero = 0.0;  // for use in dgemm
    double c0,Ival,abserr;             // results
    size_t neval;               // number of evaluations used by solver


    /* Check for proper number of arguments. */
    if ( nrhs != 4) {
        mexErrMsgIdAndTxt("MATLAB:intTwoK_mex:rhs",
            "This function requires 4 inputs.");
    }

    u = mxGetPr(prhs[0]); /* pointer to vector u */
    m = mxGetM(prhs[0]);  // dimension of problem
    w = mxGetPr(prhs[1]); // pointer to vector w
    x = mxGetPr(prhs[2]); // pointer to vector w
    sV = mxGetPr(prhs[3]);// pointer to square root matrix sV
    rank = MIN(m,n);

    // check inputs
    if (mxGetM(prhs[1]) != m || mxGetM(prhs[2]) != m || mxGetN(prhs[0]) != 1 || mxGetN(prhs[1]) != 1 || mxGetN(prhs[2]) != 1){
        mexErrMsgIdAndTxt("MATLAB:intTwoK_mex:rhs",
            "u,w,x must be [m,1].");
    }
    // check sV and put into vector sv
    if (mxGetM(prhs[3]) != m || mxGetN(prhs[3]) != m){
        mexErrMsgIdAndTxt("MATLAB:intTwoK_mex:rhs",
            "sV must be [m,m].");
        }

    // check number of outputs
    if (nlhs != 1)
    {
        mexErrMsgIdAndTxt("MATLAB:qlessqr_lapack:lhs",
            "This function requires one output matrices");
    }

    // // create output matrices
    plhs[0] = mxCreateDoubleMatrix(3,1,mxREAL);
    OUT = mxGetPr(plhs[0]);



    // put matrices together
    A = malloc(m*n*sizeof(double));
    memcpy(A,w,m*sizeof(double));
    memcpy(A+m,x,m*sizeof(double));
    memcpy(A+2*m,u,m*sizeof(double));

    // perform sV*[w,x,u]
    R = malloc(m*n*sizeof(double));
    dgemm(TRANS,TRANS,&m,&n,&m,&one,sV,&m,A,&m,&zero,R,&m);



    // now take qr of sV*[u,w,x]

    Tau = malloc(sizeof(double)*rank);  // is a size min(m,n) vector
    WORK = malloc(sizeof(double)*MAX(1,lwork));
    dgeqrf(&m,&n,R,&m,Tau,WORK,&lwork,&info); // query the routine for the optimal size of WORK
    lwork = WORK[0]; // now run the routine
    free(WORK);

    WORK = malloc(sizeof(double)*MAX(1,lwork));
    dgeqrf(&m,&n,R,&m,Tau,WORK,&lwork,&info);


    // now remember there is shit stored below the diagonal
    /* function to be integrated 
    f4 = @(t,s) erf((-s*xbar(1)+t*wbar(1)+ubar(1))/sqrt(2))...
    .*exp(-0.5*(ubar(2)-s*xbar(2)).^2);
    ifunc4 =@(s) f4(1,s)-f4(0,s);
    */

    // Calculate the constant value
    // c0 = sqrt(M_PI/2)/R[0]*exp(-0.5*R[2*m+2]*R[2*m+2]);
    c0 = M_SQRTPI*M_SQRT1_2/R[0]*exp(-0.5*R[2*m+2]*R[2*m+2]);


    // prepare for gsl

    gsl_function F;
    struct PARAMS params;
    params.ubar = R+2*m;
    params.wbar = R;
    params.xbar = R+m;

    F.function = &wrap_intfunc;
    F.params = (void *) &params;

    // memcpy(OUT,R+2*m,m*sizeof(double));
    // // OUT[0] = Ival;
    // free(WORK);
    // free(A);
    // free(R);
    // return;

    // ADAPTIVE GAUSS CONROD
    info = gsl_integration_qng(&F,0.0,1.0,sqrt(DBL_EPS),sqrt(DBL_EPS),&Ival,&abserr,&neval);
    // info = gsl_integration_qng(&F,0.0,1.0,0.0001,sqrt(DBL_EPS),&Ival,&abserr,&neval);

    // FIXED POINT
    // gsl_integration_glfixed_table * gsltab =  gsl_integration_glfixed_table_alloc ((size_t) 31);
    // Ival =  gsl_integration_glfixed (&F, 0.0, 1.0, gsltab);

    // gsl_integration_glfixed_table_free (gsltab);

    // ADAPTIVE
    // gsl_integration_workspace * gslwspace =  gsl_integration_workspace_alloc ((size_t) 3);
    // info = gsl_integration_qag (&F, 0.0, 1.0, sqrt(DBL_EPS), sqrt(DBL_EPS),(size_t)3, GSL_INTEG_GAUSS15,gslwspace, &Ival, &abserr);
    // gsl_integration_workspace_free (gslwspace);

    OUT[0] = c0*Ival;
    OUT[1] = abserr;
    OUT[2] = (double) neval;



    free(R);
    free(A);
    free(WORK);
}


double intfunc(double s,double *ubar, double *wbar, double*xbar){
    double p1,p2;

    // p1 = erf((-s*xbar[0]+wbar[0]+ubar[0])/sqrt(2))*exp(-0.5*(ubar[1]-s*xbar[1])*(ubar[1]-s*xbar[1]));
    p1 = erf((-s*xbar[0]+wbar[0]+ubar[0])*M_SQRT1_2)*exp(-0.5*(ubar[1]-s*xbar[1])*(ubar[1]-s*xbar[1]));
    // p1 = 0;
    // p2 = erf((-s*xbar[0]+ubar[0])/sqrt(2))*exp(-0.5*(ubar[1]-s*xbar[1])*(ubar[1]-s*xbar[1]));
    p2 = erf((-s*xbar[0]+ubar[0])*M_SQRT1_2)*exp(-0.5*(ubar[1]-s*xbar[1])*(ubar[1]-s*xbar[1]));
    // p2 = 0;


    return p1 - p2;
}

double wrap_intfunc(double s,void * gsl_params){
    struct PARAMS * params;
    params = (struct PARAMS *) gsl_params;


    return intfunc(s,params->ubar,params->wbar,params->xbar);
}

